from .glyph import Glyph
from .lex import Lexer
from .grammar import MathParser
from .node.writer import tostring

# inspired by
# https://github.com/Python-Markdown/markdown/blob/master/markdown/core.py
# https://github.com/pygments/pygments/blob/master/pygments/__init__.py


class MuMath:
    def __init__(self, glyph, lexer, parser):
        self.glyph = glyph
        self.lexer = lexer
        self.parser = parser

    @classmethod
    def from_area(cls, area, **options):
        glyph = Glyph.from_area(area)
        lexer = Lexer.from_glyph(glyph)
        parser = MathParser(glyph, **options)

        return cls(glyph, lexer, parser)

    def convert(self, source):
        tree = self.build(source)
        html = tostring(tree.getroot())
        return html

    def convert_file(self, file, output=None, encoding=None):
        encoding = "utf-8" if encoding is None else encoding

        with open(file, mode="r", encoding=encoding) as f:
            html = self.convert(f.read())

        if output is not None:
            with open(output, mode="w", encoding=encoding) as f:
                f.write(html)
        else:
            print(html)

        return self

    def build(self, source, root=None):
        tokens = list(self.lexer.tokenize(source))
        return self.parser.parse(tokens, root)



import markdown
import re
import xml.etree.ElementTree as ET
from markdown.blockprocessors import BlockProcessor
from markdown.inlinepatterns import InlineProcessor
from markdown.extensions import Extension
from .util import peek, pop
from .glyph import Glyph
from .lex import Lexer
from .grammar import MathParser
from itertools import chain


# Copied from util.xml
# python-markdown doesn't support Text() nodes, and uses tag.lower(), so
# we need to convert back to normal style ET.Elements.
def elify(node):
    el = ET.Element(node.tag, node.attrib)

    start = 0
    if len(node) and node[0].tag is str:
        el.text = node[0].text
        start = 1

    for cnode in node[start:]:
        if cnode.tag is str:
            # assumes text nodes aren't consecutive
            el[-1].tail = cnode.text
        else:
            el.append(elify(cnode))

    return el


def _parse_attributes(attributes):
    pattern = r"""
        (?P<attrib>[\w:!.#-]+)           # Attribute
        (?:\s*=\s*                       # Value assignment
            (?:[\'"](?P<quoted>.*)[\'"]  # Quoted value
            |(?P<unquoted>\S*))          # Unquoted value
        )?                               # Value is optional
    """

    matches = re.finditer(pattern, attributes, re.VERBOSE)

    for m in matches:
        if m.lastgroup == "attrib":
            yield (m["attrib"], None)
        elif m.lastgroup == "quoted":
            yield (m["attrib"], m["quoted"])
        elif m.lastgroup == "unquoted":
            yield (m["attrib"], m["unquoted"])


def _alias_attributes(attributes):
    LANG_ALIAS = {
        "py": "python3",
        "js": "javascript",
        "chem": "chemistry",
    }

    ATT_ALIAS = {
        "#": "linenos",
        "!": "hl_lines",
        "@": "lineanchors",
    }

    # first attribute is positional language parameter
    (att, val), attributes = peek(attributes, default=("#", None))
    if not att.startswith("#") and val is None:
        pop(attributes)
        yield ("area", LANG_ALIAS.get(att, att))

    for att, val in attributes:
        yield (ATT_ALIAS.get(att, att), val)


def _options(attributes):
    option_keys = {
        "area", "wrap", "linenos", "linenostart", "hl_lines",
        "lineanchor", "multiline", "numbering", "align",
        "display", "counter",
    }

    options = {}
    attribs = {}
    for att, val in attributes:
        (options if att in option_keys else attribs)[att] = val

    if "linenos" in options:
        counter = options.pop("linenos")
        options["counter"] = True if counter is None else int(counter)

    return options, attribs


def options_attributes(attributes):
    return _options(_alias_attributes(_parse_attributes(attributes)))


class MuMathProcessor(BlockProcessor):
    RE_FENCE_START = r'^ *(\${2,})'
    RE_FENCE_END = r''

    def test(self, parent, block):
        return re.match(self.RE_FENCE_START, block)

    def _capture_content(self, blocks):
        start_block = blocks[0]

        m = self.test(None, blocks[0])
        blocks[0] = start_block[m.end(0):]

        fence = re.escape(m.group(1))
        for i, block in enumerate(blocks):
            if n := re.search(fence, block):
                end_block = blocks[i]
                before, after = end_block[:n.start(0)], end_block[n.end(0):]

                # This reverts the blocking
                content = "\n\n".join(chain(blocks[0:i], [before]))

                blocks[i] = after

                remove_end = i + int(after.isspace())
                del blocks[:remove_end]
                return content

        # Undo (order is important)
        blocks[0] = start_block

        return None

    def run(self, parent, blocks):
        content = self._capture_content(blocks)

        if content is None:
            return False

        if content.startswith("\\"):
            # since $$\begin{align} is common
            # though \begin{align} isn't really supported
            options = ""
        else:
            options, _, content = content.partition("\n")

        options, attributes = options_attributes(options)

        glyph = Glyph.from_area(options.pop("area", None))
        lexer = Lexer.from_glyph(glyph)
        parser = MathParser(glyph, align=True, **options)

        tokens = list(lexer.tokenize(content))
        root = parser.parse(tokens).getroot()
        root.attrib.update(attributes)

        parent.append(elify(root))


class InlineMuMathProcessor(InlineProcessor):
    def handleMatch(self, m, data):
        content = m.group(1)

        if content.startswith("{"):
            # this leaves the first { intact
            # but it is ignored in the parser
            options, _, content = content.partition("}")
        else:
            options = ""
        options, attributes = options_attributes(options)

        glyph = Glyph.from_area(options.pop("area", None))
        lexer = Lexer.from_glyph(glyph)
        parser = MathParser(glyph, align=False, **options)

        tokens = list(lexer.tokenize(content))
        root = elify(parser.parse(tokens).getroot())
        root.attrib.update(attributes)

        return root, m.start(0), m.end(0)


class InlineMuMark(Extension):
    config = {
        "priority": [110, "Extension priority"],
    }

    def extendMarkdown(self, md):
        MATH_PATTERN = r"\$(.*?)\$"
        md.inlinePatterns.register(
            InlineMuMathProcessor(MATH_PATTERN, md), "mumath-inline", 110)


class MuMark(Extension):
    config = {
        "priority": [110, "Extension priority"],
    }

    def extendMarkdown(self, md):
        md.parser.blockprocessors.register(
            MuMathProcessor(md.parser), "mumath", self.getConfig("priority", 110))



def normal(): return dict(mathvariant="normal")


_ELEMENTS = [
	(1, r"H", "Helium"),     # ^1H
    (1, r"D", "Deuterium"),  # ^2H
	(1, r"T", "Tritium"),    # ^3H
    (2, r"He", "Helium"),
    (3, r"Li", "Lithium"),
    (4, r"Be", "Beryllium"),
    (5, r"B", "Boron"),
    (6, r"C", "Carbon"),
    (7, r"N", "Nitrogen"),
    (8, r"O", "Oxygen"),
    (9, r"F", "Fluorine"),
    (10, r"Ne", "Neon"),
    (11, r"Na", "Sodium"),
    (12, r"Mg", "Magnesium"),
    (13, r"Al", "Aluminium"),
    (14, r"Si", "Silicon"),
    (15, r"P", "Phosphorus"),
    (16, r"S", "Sulfur"),
    (17, r"Cl", "Chlorine"),
    (18, r"Ar", "Argon"),
    (19, r"K", "Potassium"),
    (20, r"Ca", "Calcium"),
    (21, r"Sc", "Scandium"),
    (22, r"Ti", "Titanium"),
    (23, r"V", "Vanadium"),
    (24, r"Cr", "Chromium"),
    (25, r"Mn", "Manganese"),
    (26, r"Fe", "Iron"),
    (27, r"Co", "Cobalt"),
    (28, r"Ni", "Nickel"),
    (29, r"Cu", "Copper"),
    (30, r"Zn", "Zinc"),
    (31, r"Ga", "Gallium"),
    (32, r"Ge", "Germanium"),
    (33, r"As", "Arsenic"),
    (34, r"Se", "Selenium"),
    (35, r"Br", "Bromine"),
    (36, r"Kr", "Krypton"),
    (37, r"Rb", "Rubidium"),
    (38, r"Sr", "Strontium"),
    (39, r"Y", "Yttrium"),
    (40, r"Zr", "Zirconium"),
    (41, r"Nb", "Niobium"),
    (42, r"Mo", "Molybdenum"),
    (43, r"Tc", "Technetium"),
    (44, r"Ru", "Ruthenium"),
    (45, r"Rh", "Rhodium"),
    (46, r"Pd", "Palladium"),
    (47, r"Ag", "Silver"),
    (48, r"Cd", "Cadmium"),
    (49, r"In", "Indium"),
    (50, r"Sn", "Tin"),
    (51, r"Sb", "Antimony"),
    (52, r"Te", "Tellurium"),
    (53, r"I", "Iodine"),
    (54, r"Xe", "Xenon"),
    (55, r"Cs", "Caesium"),
    (56, r"Ba", "Barium"),
    (57, r"La", "Lanthanum"),
    (58, r"Ce", "Cerium"),
    (59, r"Pr", "Praseodymium"),
    (60, r"Nd", "Neodymium"),
    (61, r"Pm", "Promethium"),
    (62, r"Sm", "Samarium"),
    (63, r"Eu", "Europium"),
    (64, r"Gd", "Gadolinium"),
    (65, r"Tb", "Terbium"),
    (66, r"Dy", "Dysprosium"),
    (67, r"Ho", "Holmium"),
    (68, r"Er", "Erbium"),
    (69, r"Tm", "Thulium"),
    (70, r"Yb", "Ytterbium"),
    (71, r"Lu", "Lutetium"),
    (72, r"Hf", "Hafnium"),
    (73, r"Ta", "Tantalum"),
    (74, r"W", "Tungsten"),
    (75, r"Re", "Rhenium"),
    (76, r"Os", "Osmium"),
    (77, r"Ir", "Iridium"),
    (78, r"Pt", "Platinum"),
    (79, r"Au", "Gold"),
    (80, r"Hg", "Mercury"),
    (81, r"Tl", "Thallium"),
    (82, r"Pb", "Lead"),
    (83, r"Bi", "Bismuth"),
    (84, r"Po", "Polonium"),
    (85, r"At", "Astatine"),
    (86, r"Rn", "Radon"),
    (87, r"Fr", "Francium"),
    (88, r"Ra", "Radium"),
    (89, r"Ac", "Actinium"),
    (90, r"Th", "Thorium"),
    (91, r"Pa", "Protactinium"),
    (92, r"U", "Uranium"),
    (93, r"Np", "Neptunium"),
    (94, r"Pu", "Plutonium"),
    (95, r"Am", "Americium"),
    (96, r"Cm", "Curium"),
    (97, r"Bk", "Berkelium"),
    (98, r"Cf", "Californium"),
    (99, r"Es", "Einsteinium"),
    (100, r"Fm", "Fermium"),
    (101, r"Md", "Mendelevium"),
    (102, r"No", "Nobelium"),
    (103, r"Lr", "Lawrencium"),
    (104, r"Rf", "Rutherfordium"),
    (105, r"Db", "Dubnium"),
    (106, r"Sg", "Seaborgium"),
    (107, r"Bh", "Bohrium"),
    (108, r"Hs", "Hassium"),
    (109, r"Mt", "Meitnerium"),
    (110, r"Ds", "Darmstadtium"),
    (111, r"Rg", "Roentgenium"),
    (112, r"Cn", "Copernicium"),
    (113, r"Nh", "Nihonium"),
    (114, r"Fl", "Flerovium"),
    (115, r"Mc", "Moscovium"),
    (116, r"Lv", "Livermorium"),
    (117, r"Ts", "Tennessine"),
    (118, r"Og", "Oganesson"),
]


# We sort by symbol name length so He is captured before H
def sym(t): return len(t[1])

identifiers = {
    r"\alembic": "⚗",  # &#x2697;
    r"\atom": "⚛",  # &#x269b;
    r"\radioactive": "☢",  # &#x2622;
    r"\biohazard": "☣",  # &#x2623;
    r"\poisonold": "☠",  # &#x2620;
    r"\equilibrium": "⇌",  # &Equilibrium;
    r"\reverseequilibrium": "⇋",  # &ReverseEquilibrium;
    r"\biequation": "⇄",  # &rightleftarrows;
    r"\requation": "→",  # &rightarrow;
    r"\Requation": "⟶",  # &longrightarrow;
    r"\lequation": "←",  # &leftarrow;
    r"\Lequation": "⟵",  # &longleftarrow;
    r"\aqua": "q",  # q
    r"\liquid": "l",  # l
    r"\gas": "g",  # g
    r"\solid": "s",  # s
    r"\togas": "↑",  # &uparrow;
    r"\tosolid": "↓",  # &downarrow;
}

for atomic_number, symbol, name in sorted(_ELEMENTS, key=sym, reverse=True):
    identifiers[symbol] = identifiers[name] = (symbol, normal())



from itertools import chain
from . import Node, Collection, Empty


def element_clone(el):
    new = Element(el.tag, attrib=el.attrib)
    new.text = el.text
    new.extend(clone(child) for child in el)
    return new


def clone(el):
    # copy.deepcopy doesn't copy type
    if isinstance(el, Node):
        return Node(el.tag, el.text, attrib=el.attrib)
    elif isinstance(el, Collection):
        children = (clone(child) for child in el)
        return Collection(el.tag, children, attrib=el.attrib)
    elif isinstance(el, Empty):
        return Empty(el.tag, attrib=el.attrib)
    else:
        raise TypeError


def traverse(el):
    # you can also do yield from el.iter(), but this is more fun
    yield el
    yield from chain.from_iterable(map(traverse, el))



import xml.etree.ElementTree as ET
import re
import io
from typing import NamedTuple
from itertools import chain
from .context.tokens import tok_name

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache

# fun equivalent
# def TokenInfo(type: str, string: str, start: str, end: int, line: int): pass
# TokenInfo = NamedTuple(TokenInfo.__name__, TokenInfo.__annotations__)

# shamelessly copied from python.tokenize, with an added flag
_TokenInfo = NamedTuple("TokenInfo", [
    ("type", int),
    ("variant", int),
    ("string", str),
    ("start", int),
    ("end", int),
    ("line", int),
    ("flag", int),
])


# Match words without _ and numbers
_keyword = re.compile(r"\\[^\W\d_]+")


class TokenInfo(_TokenInfo):
    def __repr__(self):
        return ('TokenInfo(type=%s, variant=%s, string=%r, start=%r, end=%r, line=%r, flag=%s)' %
                self._replace(type=tok_name[self.type]))


def _token_error_msg(line, pos):
    return "\n".join(["Unknown token:", line, " "*pos + "^"*(len(line) - pos)])


class Lexer:
    def __init__(self, symbols, keywords={}, flags={}):
        self.symbols = symbols
        self.types = list(symbols.values())
        self.keywords = keywords
        self.flags = flags

        self.pattern = re.compile("|".join(
            f"({pattern})" for pattern in symbols
        ), flags=re.UNICODE)

    @classmethod
    @cache
    def from_glyph(cls, glyph):
        symbols = {_keyword.pattern: None}
        keywords = {}
        for ijk, ttype, pattern in glyph.patterns():
            if _keyword.fullmatch(pattern):
                keywords[pattern] = (ttype, ijk)
            else:
                symbols[pattern] = (ttype, ijk)

        return cls(symbols, keywords, glyph.flags)

    def _tokenize(self, stream):
        flag = 0
        for lineno, line in enumerate(stream, start=1):
            pos = 0
            while m := self.pattern.match(line, pos):
                # lastindex is 1-indexed
                tindex, token, start, end =\
                    m.lastindex, m.group(), m.start(), m.end()

                if token in self.keywords:
                    ttype, variant = self.keywords[token]
                else:
                    ttype, variant = self.types[tindex-1]

                if setter := self.flags.get(ttype):
                    flag = setter(flag)
                    pos = end
                    continue

                token = TokenInfo(ttype, variant, token, start, end, lineno, flag)

                yield token
                pos = end

                # if start == end: raise ValueError("Zero-width token")
            if pos != len(line):
                raise ValueError(_token_error_msg(line, pos))

    def tokenize(self, stream):
        if isinstance(stream, str):
            stream = io.StringIO(stream)
        return self._tokenize(stream)



top_level = [
    "math",
]

tokens = [
    "mi",
    "mn",
    "mo",
    "ms",
    "mspace",
    "mtext",
]

layout = [
    "menclose",
    "merror",
    "mfenced",
    "mfrac",
    "mpadded",
    "mphantom",
    "mroot",
    "mrow",
    "msqrt",
    "mstyle",
]

script_limit = [
    "mmultiscripts",
    "mover",
    "mprescripts",
    "msub",
    "msubsup",
    "msup",
    "munder",
    "munderover",
    "none",
]

tabular = [
    "maligngroup",
    "malignmark",
    "mtable",
    "mtd",
    "mtr",
]

elementary = [
    "mlongdiv",
    "mscarries",
    "mscarry",
    "msgroup",
    "msline",
    "msrow",
    "mstack",
]

uncategorized = [
    "maction",
]

semantic = [
    "annotation",
    "annotation-xml",
    "semantics",
]


node = [
    "mi",
    "mn",
    "mo",
    "ms",
    "mtext",
]

collection = [
    "math",
    "menclose",
    "merror",
    "mfenced",
    "mfrac",
    "mpadded",
    "mphantom",
    "mroot",
    "mrow",
    "msqrt",
    "mstyle",
    "mmultiscripts",
    "mover",
    "msub",
    "msubsup",
    "msup",
    "munder",
    "munderover",
    "maligngroup",
    "malignmark",
    "mtable",
    "mtd",
    "mtr",
]

empty = [
    "none",
    "mprescripts",
    "mspace",
]



import http.server
import socketserver
import importlib
import sys
import html
from functools import partial
from http.server import HTTPStatus
from pathlib import Path
from urllib.parse import unquote
from .core import MuMath
from .util import listify

# Most of this is copied from http.server:
# https://github.com/python/cpython/blob/3.9/Lib/http/server.py

# if __name__ == "__main__" and __package__ is None:
#     __package__ = "mumath2"


class MuMathHandler(http.server.SimpleHTTPRequestHandler):
    # __init__ is socketserver.BaseRequestHandler.__init__
    # __init__ is http.server.SimpleHTTPRequestHandler.__init__


    def __init__(self, *args, case=None, area=None, **kwargs):
        self.case = case
        self.area = listify(area)
        super().__init__(*args, **kwargs)

    # @property
    # def _is_test(self):
    #     return self._path.name.startswith("mumath")

    @property
    def _path(self):
        return Path(self.translate_path(self.path))

    def _set_headers(self, content):
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-type", "text/html")
        self.send_header("Content-Length", len(content))
        # self.send_header("Last-Modified",
        #     self.date_time_string(fs.st_mtime))
        self.end_headers()

    def _set_content(self, content):
        self.wfile.write(content)

    def _get_post_data(self):
        # https://stackoverflow.com/questions/28884947/is-there-a-function-in-python-to-convert-html-entities-to-percent-encoding
        def split(string):
            key, value = string.split("=", maxsplit=1)
            return key, unquote(value)

        content_length = int(self.headers['Content-Length'])
        post_data = self.rfile.read(content_length)
        post_data = post_data.replace(b"+", b" ")
        post_data = post_data.decode("utf-8")

        parameters = dict(map(split, post_data.split("&")))

        return parameters

    def _update(self, content, data):
        # hacky way to force reload
        # for mod in list(sys.modules):
        #     if mod.startswith("mumath"):
        #         del sys.modules[mod]
        # # importlib.reload(mumath)

        if data == "":
            # data = "\n".join(tests)
            data = "\n"  # str.splitlines doesn't like empty strings

        def apply(func):
            def _apply(lines):
                return "<br>".join(map(func, lines.splitlines()))
            return _apply

        mu = MuMath.from_area(self.area)
        image = apply(mu.convert)(data)
        source = html.escape(image)

        content = content.replace(b"$INPUT", data.encode("utf-8"))
        content = content.replace(b"$OUTPUT", image.encode("utf-8"))
        content = content.replace(b"$SOURCE", source.encode("utf-8"))

        return content

    def _get_content(self, data=""):
        path = Path(__file__).parent / "data" / "index.html"

        content = path.read_bytes()
        return self._update(content, data)

    def do_GET(self):
        if self._path.name == "favicon.ico":
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")
            return

        content = self._get_content(case)

        self._set_headers(content)
        self._set_content(content)


        # path = Path(self.translate_path(self.path))

        # print(path)

        # path = Path(self.translate_path(self.path))
        # if path.is_dir():
        #     path = path / "index"

        # if path.with_suffix(".add").is_file():
        #     # importlib.reload(core)
        #     from .core import Addup

        #     html = path.with_suffix(".html")
        #     addup = html.with_suffix(".add")
        #     self.log_message(f"Converting {html}")
        #     Addup().convert_file(addup, html)

        # super().do_GET()

    def do_POST(self):
        # https://stackoverflow.com/a/53348829
        post_data = self._get_post_data()
        content = self._get_content(post_data.get("mumath", "error"))

        self._set_headers(content)
        self._set_content(content)


def _get_best_family(*address):
    # pre-3.8
    import socket
    infos = socket.getaddrinfo(
        *address,
        type=socket.SOCK_STREAM,
        flags=socket.AI_PASSIVE,
    )
    family, _type, _proto, _canonname, sockaddr = next(iter(infos))
    return family, sockaddr


def serve(HandlerClass=http.server.BaseHTTPRequestHandler,
         ServerClass=http.server.ThreadingHTTPServer,
         protocol="HTTP/1.0", port=8000, bind=None, open=False):
    """Test the HTTP request handler class.
    This runs an HTTP server on port 8000 (or the port argument).
    """
    ServerClass.address_family, addr = _get_best_family(bind, port)

    HandlerClass.protocol_version = protocol
    with ServerClass(addr, HandlerClass) as httpd:
        host, port = httpd.socket.getsockname()[:2]
        url_host = f'[{host}]' if ':' in host else host
        url = f"http://{url_host}:{port}/"
        url = url.replace("[::]", "localhost")
        print(f"Serving HTTP on {host} port {port} ({url}) ...")
        if open:
            import webbrowser
            webbrowser.open_new_tab(url)

        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            print("\nKeyboard interrupt received, exiting.")
            import sys; sys.exit(0)


if __name__ == "__main__":
    import argparse
    import os
    import socket
    import contextlib
    from functools import partial

    parser = argparse.ArgumentParser()
    parser.add_argument('--cgi', action='store_true',
                       help='Run as CGI Server')
    parser.add_argument('--bind', '-b', metavar='ADDRESS',
                        help='Specify alternate bind address '
                             '[default: all interfaces]')
    parser.add_argument('--directory', '-d', default=os.getcwd(),
                        help='Specify alternative directory '
                        '[default:current directory]')
    parser.add_argument('port', action='store',
                        default=8000, type=int,
                        nargs='?',
                        help='Specify alternate port [default: 8000]')

    parser.add_argument('--open', action='store_true',
                        help='Open in new tab')
    parser.add_argument('--area', '-a', nargs='*',
                        help='Select dictionary')
    mexparser = parser.add_mutually_exclusive_group()
    mexparser.add_argument('--example', action='store_true',
                        help='Run example cases')
    mexparser.add_argument('--equation', '-e',
                        help='Pass equation as argument')
    mexparser.add_argument('--file', '-f',
                        help='Open file')

    args = parser.parse_args()
    if args.cgi:
        handler_class = http.server.CGIHTTPRequestHandler
    else:
        if args.example:
            case = "e^\\tau = 1\nI = \\matrix[1,0;0,1]"
        elif args.equation:
            case = args.equation
        elif args.file:
            case = Path(args.file).read_text(encoding="utf-8")
        else:
            case = ""

        handler_class = partial(
            MuMathHandler,
            directory=args.directory,
            case=case,
            area=args.area,
        )

    # ensure dual-stack is not disabled; ref #38907
    class DualStackServer(http.server.ThreadingHTTPServer):
        def server_bind(self):
            # suppress exception when protocol is IPv4
            with contextlib.suppress(Exception):
                self.socket.setsockopt(
                    socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 0)
            return super().server_bind()

    # print(args.directory)

    serve(
        HandlerClass=handler_class,
        ServerClass=DualStackServer,
        port=args.port,
        bind=args.bind,
        open=args.open,
    )





from .context.tokens import *
from .context import base
from .util import listify
from importlib import import_module
from collections import namedtuple
from types import ModuleType, SimpleNamespace

try:
    from functools import cache
except ImportError:
    from functools import lru_cache as cache


# For diffs
# The functions look weird, but they set the first and second bit
# to either 0 or 1. E.g. `(+1).__or__(0b00) == 0b01`.
flags = {
    OPEN_NEXT: (+1).__or__,
    SHUT_NEXT: (~1).__and__,
    OPEN_PREV: (+2).__or__,
    SHUT_PREV: (~2).__and__,
}


ttypes = {
    # basic syntactical components
    # KEYWORD: "keyword",
    TEXT_SEP: "text_separator",
    SUBB: "subb",
    SUB: "sub",
    SUPP: "supp",
    SUP: "sup",
    OPEN_NEXT: "open_next",
    SHUT_NEXT: "shut_next",
    OPEN_PREV: "open_prev",
    SHUT_PREV: "shut_prev",
    SOFT_SPACE: "soft_space",
    STRING: "string",
    COMMENT: "comment",

    # basic functional components
    MATRIX: "matrix",
    BEGIN: "begin",
    END: "end",
    ENVIRONMENT: "environments",
    OVER: "over",
    CHOOSE: "choose",
    SERIES: "series",  # macro
    SQRT: "sqrt",
    CLASS_: "class_",
    TEXT: "text",
    NO_NUMBER: "no_number",
    PRESCRIPT: "prescript",
    UNDERSET: "underset",
    OVERSET: "overset",
    FRAC: "frac",
    BINOM: "binom",
    ROOT: "root",
    DISPLAYSTYLE: "displaystyle",
    PAD: "pad",

    # general components
    RELATION: "relations",
    IDENTIFIER: "identifiers",
    OPERATOR: "operators",
    BINOP: "binary_operators",
    FUNCTION: "functions",
    NORM: "brackets",

    NUMBER: "numbers",
    HAT: "hats",
    SHOE: "shoes",
    OPEN: "open_brackets",
    CLOSE: "close_brackets",
    COL_SEP: "col_separators",
    ROW_SEP: "row_separators",
    VARIANT: "fonts",
    ENCLOSE: "enclosures",
    SPACE: "spaces",

    # Must come last so it doesn't override custom words
    # Maybe separate type?
    WORD: "words",
}


# May add a "name" to this
Symbol = namedtuple("Symbol", ("pattern", "property"))


class Context(namedtuple("Context", ttypes.values())):
    ttypes = ttypes

    @staticmethod
    def symbols(symbols):
        return list(map(Symbol._make, symbols.items()))

    @classmethod
    def from_base(cls):
        glyph = cls._make(cls.symbols(getattr(base, name)) for name in cls._fields)
        glyph.normalize()

        return glyph

    @classmethod
    def from_namespace(cls, context):
        glyph = cls._make(cls.symbols(getattr(context, name, {})) for name in cls._fields)
        glyph.normalize()

        return glyph

    @staticmethod
    def _normalize_symbol(attrib):
        if isinstance(attrib, tuple):
            return attrib
        elif isinstance(attrib, str):
            return (attrib, {})
        else:
            raise TypeError(f"Unknown type {type(attrib)} for {attrib}")

    def _normalize(self, symbols):
        for i, symbol in enumerate(symbols):
            symbols[i] = symbol._replace(
                property=self._normalize_symbol(symbol.property))

    def normalize(self):
        for symbols in [
            self.operators,
            self.binary_operators,
            self.relations,
            self.functions,
            self.identifiers,
            self.brackets,
        ]:
            self._normalize(symbols)

    def items(self):
        # ttype_name is aligned with Glyph
        yield from zip(ttypes, self)


class Glyph:
    areas = []
    index = {}
    flags = flags

    def __init__(self, areas, index):
        # Note: instance attributes are different from class attributes
        self.areas = areas
        self.index = index

    @classmethod
    def _get(cls, area):
        # Lazy getter
        context = cls.areas[cls.index[area]]
        if not isinstance(context, Context):
            if isinstance(context, str):
                context = import_module(context)
            elif isinstance(context, ModuleType):
                pass
            elif isinstance(context, dict):
                context = SimpleNamespace(**context)
            else:
                raise TypeError(f"Invalid type '{type(context)}' for area.")

            context = Context.from_namespace(context)
            cls.areas[cls.index[area]] = context
        return context

    @classmethod
    @cache
    def from_area(cls, area=None, base=True):
        areas = listify(area)
        if base is True:
            areas = areas + ("base",)

        index = {area: cls.index[area] for area in areas}
        areas = [cls._get(area) for area in areas]

        return cls(areas, index)


    @classmethod
    def register_area(cls, area, context):
        # TODO: Handle overwrites?
        assert area not in cls.index
        cls.areas.append(context)
        cls.index[area] = cls.areas.index(context)

    @classmethod
    def register_areas(cls, area=None, **areas):
        if area is None:
            for area, context in areas.items():
                cls.register_area(area, context)
        if isinstance(area, dict):
            for area, context in area.items():
                cls.register_area(area, context)
        else:
            raise TypeError(f"Invalid type {type(area)} for area.")

    @classmethod
    def register_extensions(cls):
        # from importlib import metadata
        # metadata.entry_points(group="mumath.extensions")
        pass

    def __iter__(self):
        for i, area in enumerate(self.areas):
            for j, (ttype, symbols) in enumerate(area.items()):
                for k, symbol in enumerate(symbols):
                    yield (i, j, k), ttype, symbol

    def patterns(self):
        for ijk, ttype, symbol in self:
            yield ijk, ttype, symbol.pattern

    def __getitem__(self, key):
        k0, k1, k2 = key
        return self.areas[k0][k1][k2]

    def property(self, key, default=None):
        try:
            return self[key].property
        except KeyError:
            return default


# This should probably be defined at instance-level
# But that requires some refactoring
Glyph.register_areas({
    "base": "mumath.context.base",
    "physics": "mumath.context.physics",
    "chemistry": "mumath.context.chemistry",
    "statistics": "mumath.context.statistics",
})
Glyph.register_extensions()



from collections.abc import Iterable
from itertools import chain


def listify(var):
    if var is None:
        return ()
    elif isinstance(var, str):
        return tuple(var.split(","))
    elif isinstance(var, Iterable):
        return tuple(var)
    else:
        raise TypeError


def peek(it, default=None):
    try:
        el = pop(it)
        return el, chain([el], it)
    except StopIteration:
        # Empty sequence
        return default, it

def pop(it):
    return next(it)



# tokens and special symbols for lexer
# based on builtins.token

IDENTIFIER = 1
NUMBER = 2
OPERATOR = 3
STRING = 4
SPACE = 5
TEXT = 6

KEYWORD = 7
BINOP = 8
RELATION = 9
COL_SEP = 10
ROW_SEP = 11
TEXT_SEP = 12
SUBB = 13
SUB = 14
SUPP = 15
SUP = 16
OPEN = 17
CLOSE = 18

OPEN_NEXT = 19
SHUT_NEXT = 20
OPEN_PREV = 21
SHUT_PREV = 22

FUNCTION = 23
MATRIX = 24
BEGIN = 25
END = 26
OVER = 27
CHOOSE = 28
LARGEOP = 29
SERIES = 30
HAT = 31
NORM = 32
SQRT = 33
VARIANT = 34
ENCLOSE = 35
CLASS_ = 36  # note _
NO_NUMBER = 37

SOFT_SPACE = 38
COMMENT = 39
PRESCRIPT = 40

UNDERSET = 41
OVERSET = 43
FRAC = 44
BINOM = 45
ROOT = 46
DISPLAYSTYLE = 47
PAD = 48
SHOE = 49
ENVIRONMENT = 50
WORD = 51

tok_name = {value: name
            for name, value in globals().items()
            if isinstance(value, int) and not name.startswith('_')}



import xml.etree.ElementTree as ET
from functools import partial
from types import SimpleNamespace as Namespace
from .mathml import node, collection, empty
import html as _html


def Text(text):
    el = ET.Element(str)
    el.text = text
    return el


class Element(ET.Element):
    def __repr__(self):
        return f"<{self._name()} />"

    def _name(self):
        name = [self.tag]
        for key, value in self.attrib.items():
            name.append(f'{key}="{value}"')
        return " ".join(name)


class Node(Element):
    def __init__(self, tag, text=None, attrib={}, **extra):
        super().__init__(tag, attrib, **extra)
        # self.text = text
        self.append(Text(text))

    def __repr__(self):
        return f"<{self._name()}>{self[0].text}</{self.tag}>"



class Collection(Element):
    def __init__(self, tag, children=(), attrib={}, **extra):
        super().__init__(tag, attrib, **extra)
        self.extend(children)

    def __repr__(self):
        return f"<{self._name()} with {len(self)} children>"


class Empty(Element):
    # For elements like <mprescripts>, <none>, <mspace>
    pass


class Comment(Element):
    def __repr__(self):
        return f"<!-- {self.text} -->"

_nodes = [
    (Node, node),
    (Collection, collection),
    (Empty, empty),
]


mml = Namespace(**{
    tag: partial(mtype, tag)
    for mtype, tags in _nodes
    for tag in tags
})

html = Namespace(Comment=Comment)

special = Namespace(NO_NUMBER=Empty("NO_NUMBER"))




from . import Node, Collection, Empty, Comment

def _open(el):
    label = [el.tag]
    for k, v in el.attrib.items():
        label.append(f'{k}="{v}"' if v != "" else k)
    return " ".join(label)


def text_content(el):
    return "".join(el.itertext())


def _write(write, el, indent, level, end):
    ind = level*indent
    if isinstance(el, Node):
        write(f"{ind}<{_open(el)}>{text_content(el)}</{el.tag}>{end}")
    elif isinstance(el, Collection):
        write(f"{ind}<{_open(el)}>{end}")
        for child in el:
            _write(write, child, indent, level+1, end)
        write(f"{ind}</{el.tag}>{end}")
    elif isinstance(el, Empty):
        write(f"{ind}<{_open(el)} />{end}")
    elif isinstance(el, Comment):
        write(f"{ind}<!-- {el.text} -->")
    else:
        raise TypeError


def tostring(el, level=0, indent="  ", end="\n"):
    import io
    stream = io.StringIO()
    _write(stream.write, el, indent, level, end)
    return stream.getvalue()


def treeprint(el, level=0, indent="  ", end="\n"):
    import sys
    _write(sys.stdout.write, el, indent, level, end)
    sys.stdout.flush()



attrib = dict
def atomic(*args): return dict.fromkeys(args)

# no functions in this file; this should be easily convertible to json/xml

# group = lambda l: re.match(r'.*(r"\\.*?").*(&.*?;|\w+)"\)', l).groups()
# fmt = lambda t: f'    {t[0]}: "{unescape(t[1])}",  # {t[1]}'
# lambda ls: "\n".join(map(fmt, map(group, filter(None, str.splitlines(ls)))))


# keyword = atomic(r"\\[^\W\d_]+")
text_separator = atomic(r"\$")
subb = atomic(r"\_\_")
sub = atomic(r"\_")
supp = atomic(r"\^\^")
sup = atomic(r"\^")
open_next = atomic(r"\\\[")
shut_next = atomic(r"\\\)")
open_prev = atomic(r"\\\(")
shut_prev = atomic(r"\\\]")
soft_space = atomic(r"\s+")
string = atomic(r'".*?"')
comment = atomic(r'\%.*$')
defaults = {
    r"\\\&": "&",
    r"\\\%": "%",
    r"\\\$": "$",
    r"\\\#": "#",
    r"\\\_": "_",
    # r"\\\\": "\\",
    r"\\\.": ".",
}


matrix = atomic(r"\matrix", r"\cases")
begin = atomic(r"\begin")
end = atomic(r"\end")
over = atomic(r"\over", r"\bover")
choose = atomic(r"\choose")
series = atomic(r"\series")  # macro
sqrt = atomic(r"\sqrt")
class_ = atomic(r"\class")
text = atomic(r"\text")
no_number = atomic(r"\nonumber", r"\notag")
prescript = atomic(r"\prescript")
underset = atomic(r"\underset")
overset = atomic(r"\overset")
frac = atomic(r"\frac")
binom = atomic(r"\binom")
root = atomic(r"\root")
displaystyle = atomic(r"\displaystyle")
pad = atomic(r"\pad")
words = atomic(r"[^\W\d_]+")


numbers = atomic(
    r"0[xX](?:_?[0-9a-fA-F])+",
    r"0[bB](?:_?[01])+",
    r"0[oO](?:_?[0-7])+",
    # decimals of the form .1 or 1. are not permitted.
    r"[0-9](?:_?[0-9])*[\.,](?:[0-9](?:_?[0-9])*)",
    r"(?:0(?:_?0)*|[1-9](?:_?[0-9])*)",
    # roman numerals
    r"(?=[MDCLXVI])M*(?:C[MD]|D?C{0,3})(?:X[CL]|L?X{0,3})(?:I[XV]|V?I{0,3})",
)


custom_identifiers = {
    r"a": "a",
    r"b": "b",
    r"c": "c",
    r"x": "x",
    r"y": "y",
    r"z": "z",
}


custom_functions = {
    r"f": ("f", attrib(form="prefix")),
    r"g": ("g", attrib(form="prefix")),
    r"h": ("h", attrib(form="prefix")),
}


identifiers = {
    r"\?": "?",
    r"\ell": ("ℓ", {}),  # &ell;
    # r"He": ("He", attrib(mathvariant="normal")),
    # r"Pb": ("Pb", attrib(mathvariant="normal")),
    # r"Tl": ("Tl", attrib(mathvariant="normal")),
    # r"H": ("H", attrib(mathvariant="normal")),
    # r"N": ("N", attrib(mathvariant="normal")),
    r"\infty": ("∞", {}),  # &infin;
    r"\aleph": "ℵ",  # &alefsym;
    r"\imath": ("ı", attrib(mathvariant="italic")),  # &imath;
    r"\jmath": ("ȷ", attrib(mathvariant="italic")),  # &jmath;

    r"\top": "⊤",  # &top;
    r"\bot": "⊥",  # &bot;
    r"\Box": "□",  # &#x25a1;
    # r"\reals": ("&reals;", attrib(mathvariant="double-struck"))

    r"\.{3}": "…",
    r"\ldots": "…",  # &hellip;
    r"\cdots": "⋯",  # &ctdot;
    r"\vdots": "⋮",  # &vellip;
    r"\ddots": "⋱",  # &dtdot;
    r"\Ddots": "⋰",  # &utdot;

    r"\prime": "′",  # &prime;
    r"\qed": "□",  # &#x25a1;

    r"\angle": "∠",  # &angle;
}

identifiers.update(custom_identifiers)

operators = {
    # r"\sum": ("∑", attrib(form="prefix", movablelimits="true", largeop="true")),
    # r"\int": ("∫", attrib(form="prefix", movablelimits="true", largeop="true")),  # not Sigma!
    r"\lim": ("lim", attrib(form="prefix", movablelimits="true")),
    r"\argmin": ("argmin", attrib(form="prefix", movablelimits="true")),
    r"\to": ("→", attrib(form="infix")),
    # r"\in": ("∈", attrib(form="infix")),
    r"\det": ("det", attrib(form="prefix", rspace="0")),
    r"\nabla": ("∇", attrib(form="prefix", rspace="0")),  # &nabla;
    r"\del": ("∇", attrib(form="prefix", rspace="0")),  # &Del;
    r"\d": ("ⅆ", attrib(form="prefix", rspace="0")),  # &DifferentialD;
    r"\partial": ("∂", attrib(form="prefix", rspace="0")),  # &PartialD;

    # r"d": ("d", attrib(form="prefix", rspace="0")),
    # r"-->": ("→", attrib(form="infix", stretchy="true")),
    r"\\\|": ("‖", attrib()),  # &Vert;
    r"\\\*": ("*", attrib()),  # &it;
    r"\\\+": ("&#8292;", attrib()),  # For mixed numbers
    r"\\\¤": ("&ApplyFunction;", attrib()),  # &af;
    r"!": ("!", attrib(form="postfix", lspace="0")),


    # large operators
    r"\sum": ("∑", attrib(form="prefix", largeop="true", movablelimits="true")),  # &sum;
    r"\prod": ("∏", attrib(form="prefix", largeop="true", movablelimits="true")),  # &prod;
    r"\coprod": ("∐", attrib(form="prefix", largeop="true", movablelimits="true")),  # &coprod;
    r"\int": ("∫", attrib(form="prefix", largeop="true")),  # &#x222B;
    r"\iint": ("∬", attrib(form="prefix", largeop="true")),  # &#x222C;
    r"\iiint": ("∭", attrib(form="prefix", largeop="true")),  # &#x222D;
    r"\oint": ("∲", attrib(form="prefix", largeop="true")),  # &#x2232;
    r"\bigcap": ("⋂", attrib(form="prefix", largeop="true", movablelimits="true")),  # &bigcap;
    r"\intersection": ("⋂", attrib(form="prefix", largeop="true", movablelimits="true")),  # &Intersection;
    r"\bigcup": ("⋃", attrib(form="prefix", largeop="true", movablelimits="true")),  # &bigcup;
    r"\union": ("⋃", attrib(form="prefix", largeop="true", movablelimits="true")),  # &Union;
    r"\bigsqcup": ("⨆", attrib(form="prefix", largeop="true", movablelimits="true")), # &bigsqcup;
    r"\bigvee": ("⋁", attrib(form="infix", largeop="true")), # &Vee;
    r"\bigwedge": ("⋀", attrib(form="infix", largeop="true")), # &Wedge;
    r"\bigodot": ("⨀", attrib(form="prefix", largeop="true", movablelimits="true")),  # &bigodot;
    r"\bigotimes": ("⨂", attrib(form="prefix", largeop="true", movablelimits="true")),  # &bigotimes;
    r"\bigoplus": ("⨁", attrib(form="prefix", largeop="true", movablelimits="true")),  # &bigoplus;
    r"\biguplus": ("⨄", attrib(form="prefix", largeop="true", movablelimits="true")),  # &biguplus;


    r"\forall": ("∀", attrib(form="prefix", largeop="true")),  # &forall;
    r"\exists": ("∃", attrib(form="prefix", largeop="true")),  # &exist;
}

operators.update(defaults)


binary_operators = {
    r"\+": "+",
    r"\-": "-",
    r"\*": "&InvisibleTimes;",
    r"\/": "/",
    r"\.": ".",

    r"\pm": "±",
    r"\mp": "∓",

    r"\times": "×",  # &times;
    r"\div": "÷",  # &div;
    r"\cross": "⨯",  # &Cross;
    r"\ast": "*",  # &ast;
    r"\star": "☆",  # &star;
    r"\circ": "∘",  # &#8728;
    r"\bullet": "•",  # &bullet;
    r"\cdot": "·",  # &centerdot;
    r"\cap": "∩",  # &cap;
    r"\cup": "∪",  # &cup;
    r"\given": "|",  # &vert;
    r"\uplus": "⊎",  # &uplus;
    r"\sqcap": "⊓",  # &sqcap;
    r"\sqcup": "⊔",  # &sqcup;
    r"\vee": "∨",  # &vee;
    r"\wedge": "∧",  # &wedge;
    r"\setminus": "∖",  # &setminus;
    r"\wr": "≀",  # &wr;
    r"\diamond": "⋄",  # &diamond;
    r"\bigtriangleup": "△",  # &bigtriangleup;
    r"\bigtriangledown": "▽",  # &bigtriangledown;
    r"\triangleleft": "⊲",  # &LeftTriangle;
    r"\triangleright": "⊳",  # &RightTriangle;
    r"\lhd": "⊲",  # &LeftTriangle;
    r"\rhd": "⊳",  # &RightTriangle;
    r"\unlhd": "⊴",  # &LeftTriangleEqual;
    r"\unrhd": "⊵",  # &RightTriangleEqual;
    r"\oplus": "⊕",  # &oplus;
    r"\ominus": "⊖",  # &ominus;
    r"\otimes": "⊗",  # &otimes;
    r"\oslash": "⊘",  # &osol;
    r"\odot": "⊙",  # &odot;
    r"\ocirc": "⊚",  # &ocir;
    r"\bigcirc": "○",  # &cir;
    r"\dagger": "†",  # &dagger;
    r"\ddagger": "‡",  # &Dagger;
    r"\amalg": "⨿",  # &#x2a3f;
    r"\bowtie": "⋈",  # &bowtie;
    r"\Join": "⋈",  # &bowtie;
    r"\ltimes": "⋉",  # &ltimes;
    r"\rtimes": "⋊",  # &rtimes;
    r"\smile": "⌣",  # &smile;
    r"\frown": "⌢",  # &frown;
}

relations = {
    r"<=": "≤",
    r">=": "≥",
    r"<=>": "⇔",
    r"==>": "⇒",
    r"<->": "⟷",
    r"==": "⩵",
    r"=": "=",
    r"<": "&lt;",  # <
    r">": "&gt;",  # >
    r":=": "≔",

	r"\eq": "=",  # &equals;
	r"\qeq": "≟", # &questeq;
	r"\gtreqless": "⋛",
	r"\lesseqgtr": "⋚",
	r"\gtrless": "≶",
	r"\lessgtr": "≷",
    r"\leq": "≤",  # &leq;
    r"\prec": "≺",  # &Precedes;
    r"\preceq": "⪯",  # &PrecedesEqual;
    r"\ll": "≪",  # &ll;
    r"\subset": "⊂",  # &subset;
    r"\subseteq": "⊆",  # &subseteq;
    r"\sqsubset": "⊏",  # &sqsubset;
    r"\sqsubseteq": "⊑",  # &sqsubseteq;
    r"\in": "∈",  # &in;
    r"\ni": "∋",  # &ni;
    r"\vdash": "⊢",  # &vdash;
    r"\vDash": "⊨",  # &vDash;
    r"\geq": "≥",  # &geq;
    r"\succ": "≻",  # &Succeeds;
    r"\succeq": "⪰",  # &SucceedsEqual;
    r"\gg": "≫",  # &gg;
    r"\supset": "⊃",  # &supset;
    r"\supseteq": "⊇",  # &supseteq;
    r"\sqsupset": "⊐",  # &sqsupset;
    r"\sqsupseteq": "⊒",  # &sqsupseteq;
    r"\dashv": "⊣",  # &dashv;
    r"\Dashv": "⫤",  # &Dashv;
    r"\equiv": "≡",  # &equiv;
    r"\sim": "∼",  # &sim;
    r"\simeq": "≃",  # &simeq;
    r"\asymp": "≍",  # &asympeq;
    r"\approx": "≈",  # &approx;
    r"\cong": "≅",  # &cong;
    r"\doteq": "≐",  # &doteq;
    r"\propto": "∝",  # &prop;
    r"\models": "⊧",  # &models;
    r"\perp": "⟂",  # &#x27c2;
    r"\mid": "∣",  # &mid;
    r"\parallel": "∥",  # &parallel;
    r"\implies": "⇒",  # &Implies;
    r"\iff": "⇔",  # &Leftrightarrow;
    r"\equivalently": "⇔",  # &Leftrightarrow;
    r"\mapsto": "↦",  # &mapsto;
    # r"\to": "→",  # &rightarrow;
    r"\longmapsto": "⟼",  # &longmapsto;
    r"\leadsto": "⤳",  # &rarrc;

    # r"\not" should be applied to a relation to get not_relation
    r"/=": "≠",
    r"!=": "≠",

	r"\neg": "¬",  # &not;  # differs from mm 1.0
	r"\neq": "≠", # "&NotEqual;"
	r"\ngtrless": "≸",
	r"\nlessgtr": "≹",

    r"\nleq": "≮",  # &NotLess;
    r"\nprec": "⊀",  # &NotPrecedes;
    r"\npreceq": "⋠",  # &NotPrecedesSlantEqual;
	# <Not much less than> would be here
    r"\nsubset": "⊄",  # &nsub;
    r"\nsubseteq": "⊈",  # &nsubseteq;
    r"\nsqsubset": "⊏̸",  # &NotSquareSubset;
    r"\nsqsubseteq": "⋢",  # &NotSquareSubsetEqual;
    r"\nin": "∉",  # &notin;
    r"\nni": "∌",  # &notni;
    r"\nvdash": "⊬",  # &nvdash;
    r"\nvDash": "⊭",  # &nvDash;
    r"\ngeq": "≯",  # &NotGreater;
    r"\nsucc": "⊁",  # &NotSucceeds;
    r"\nsucceq": "⋡",  # &NotSucceedsSlantEqual;
	# <Not much greater than> would be here
    r"\nsupset": "⊅",  # &nsup;
    r"\nsupseteq": "⊉",  # &nsupseteq;
    r"\nsqsupset": "⊐̸",  # &NotSquareSuperset;
    r"\nsqsupseteq": "⋣",  # &NotSquareSupersetEqual;
    r"\ndashv": "&ndashv;",  # &ndashv;
	#<nDashv> would be here
	#<nequiv> would be here
    r"\nsim": "≁",  # &nsim;
    r"\nsimeq": "≄",  # &nsimeq;
    r"\nasymp": "≭",  # &NotCupCap;
    r"\napprox": "≉",  # &napprox;
    r"\ncong": "≇",  # &ncong;
	#<ndoteq> would be here
	#<npropto> would be here
	#<nmodels> would be here
	#<nperp> would be here
    r"\nmid": "∤",  # &nmid;
    r"\nparallel": "∦",  # &nparallel;

    r"\nimplies": "⇏",  # &nRightarrow;
    r"\niff": "⇎",  # &nLeftrightarrow;
    r"\nequivalently": "⇎",  # &nLeftrightarrow;

    # arrows
    r"-->": "→",

    r"\leftarrow": "←",  # &leftarrow;
    r"\Leftarrow": "⇐",  # &Leftarrow;
    r"\twoheadleftarrow": "↞",  # &twoheadleftarrow;
    r"\rightarrow": "→",  # &rightarrow;
    r"\Rightarrow": "⇒",  # &Rightarrow;
    r"\twoheadrightarrow": "↠",  # &twoheadrightarrow;
    r"\leftrightarrow": "↔",  # &leftrightarrow;
    r"\Leftrightarrow": "⇔",  # &Leftrightarrow;
    r"\hookleftarrow": "↩",  # &hookleftarrow;
    r"\leftharpoonup": "↼",  # &leftharpoonup;
    r"\leftharpoondown": "↽",  # &leftharpoondown;
    r"\rightleftharpoons": "⇌",  # &rightleftharpoons;
    r"\longleftarrow": "⟵",  # &longleftarrow;
    r"\Longleftarrow": "⟸",  # &Longleftarrow;
    r"\longrightarrow": "⟶",  # &longrightarrow;
    r"\Longrightarrow": "⟹",  # &Longrightarrow;
    r"\longleftrightarrow": "⟷",  # &longleftrightarrow;
    r"\Longleftrightarrow": "⟺",  # &Longleftrightarrow;
    r"\hookrightarrow": "↪",  # &hookrightarrow;
    r"\righttharpoonup": "⇀",  # &rightharpoonup;
    r"\rightharpoondown": "⇁",  # &rightharpoondown;
    r"\uparow": "↑",  # &uparrow;
    r"\Uparrow": "⇑",  # &Uparrow;
    r"\downarrow": "↓",  # &downarrow;
    r"\Downarrow": "⇓",  # &Downarrow;
    r"\updownarrow": "↕",  # &updownarrow;
    r"\Updownarrow": "⇕",  # &Updownarrow;
    r"\nearrow": "↗",  # &nearrow;
    r"\searrow": "↘",  # &searrow;
    r"\swarrow": "↙",  # &swarrow;
    r"\nwarrow": "↖",  # &nwarrow;
}

functions = {
    r"\arg": ("arg", attrib(form="prefix")),
    r"\deg": ("deg", attrib(form="prefix")),
    r"\cos": ("cos", attrib(form="prefix")),
    r"\cosh": ("cosh", attrib(form="prefix")),
    r"\sin": ("sin", attrib(form="prefix")),
    r"\sinh": ("sinh", attrib(form="prefix")),
    r"\tan": ("tan", attrib(form="prefix")),
    r"\tanh": ("tanh", attrib(form="prefix")),
    r"\exp": ("exp", attrib(form="prefix")),
    r"\log": ("log", attrib(form="prefix")),
    r"\lg": ("lg", attrib(form="prefix")),
    r"\ln": ("ln", attrib(form="prefix")),
    r"\lim": ("lim", attrib(form="prefix", movablelimits="true")),
    r"\sup": ("sup", attrib(form="prefix", movablelimits="true")),
    r"\limsup": ("limsup", attrib(form="prefix", movablelimits="true")),
    r"\inf": ("inf", attrib(form="prefix", movablelimits="true")),
    r"\liminf": ("liminf", attrib(form="prefix", movablelimits="true")),
    r"\max": ("max", attrib(form="prefix", movablelimits="true")),
    r"\argmax": ("argmax", attrib(form="prefix", movablelimits="true")),
    r"\min": ("min", attrib(form="prefix", movablelimits="true")),
    r"\argmin": ("argmin", attrib(form="prefix", movablelimits="true")),
    r"\det": ("det", attrib(form="prefix")),
    r"\diag": ("diag", attrib(form="prefix")),
    r"\ker": ("ker", attrib(form="prefix")),
    r"\mod": ("mod", attrib(form="prefix")),
    r"\sgn": ("sgn", attrib(form="prefix")),
    r"\fourier": ("ℱ", attrib(form="prefix")), # &Fouriertrf;
    r"\laplace": ("ℒ", attrib(form="prefix")),  # &Laplacetrf;
    r"\mellin": ("ℳ", attrib(form="prefix")),  # &Mellintrf;
}

operators.update(custom_functions)

hats = {
    r"\\\~": "&tilde;",
    r"\\\^": "^",
    r"\hat": "^",
	r"\check": "&check;",
	r"\acute": "&acute;",
	r"\grave": "&grave;",
	r"\bar": "&horbar;",
	r"\vec": "&rarr;", # "&#8407;"),
	r"\dot": "&dot;",
	r"\ddot": "&Dot;",
	r"\breve": "&breve;",
	r"\tilde": "&tilde;",
	r"\overline": "&oline;",
		r"\ov": "&OverBar;",
		r"\inverse": "&macr;",
	r"\underline": "&uline;",
	r"\overbrace": "&OverBrace;",
}

shoes = {
	r"\underbrace": "&UnderBrace;",
}


environments = {
    r"align":    (None, None),
    r"equation": (None, None),

    r"matrix":  (None, None),
    r"pmatrix": ("(", ")"),
    r"bmatrix": ("[", "]"),
    r"Bmatrix": ("{", "}"),
    r"vmatrix": ("|", "|"),
    r"Vmatrix": ("‖", "‖"),
    r"cases":   ("{", None),
}


brackets = {
    r"\abs":   ("|", "|"),
    r"\norm":  ("‖", "‖"),
    r"\inner": ("⟨", "⟩"),
    r"\ceil":  ("⌈", "⌉"),
    r"\floor": ("⌊", "⌋"),
    r"\round": ("⌊", "⌉"),
}


open_brackets = {
    r"\left": None,
    r"\{": None,

    r"\(":         "(",
    r"\lparent":   "(",
    r"\[":         "[",
    r"\lbracket":  "[",
    r"\\\{":       "{",
    r"\lbrace":    "{",
	r"\lfloor":    "&lfloor;",
	r"\lceil":     "&lceil;",
	r"\langle":    "&langle;",
	r"\lvert":     "&vert;",
	r"\lVert":     "&Vert;",
	r"\lucorner":  "&#11810;",
	r"\lbcorner":  "&#11812;",
}


close_brackets = {
    r"\right": None,
    r"\}": None,

    r"\)":         ")",
    r"\rparent":   ")",
    r"\]":         "]",
    r"\rbracket":  "]",
    r"\\\}":       "}",
    r"\rbrace":    "}",
	r"\rfloor":    "&rfloor;",
	r"\rceil":     "&rceil;",
	r"\rangle":    "&rangle;",
	r"\rvert":     "&vert;",
	r"\rVert":     "&Vert;",
	r"\rucorner":  "&#11811;",
	r"\rbcorner":  "&#11813;",
}

col_separators = {
    r"\middle": None,

    r"\|": "|",
    r":": ":",
    r"\\,": "&InvisibleComma;",  # &ic;
    r",": ",",
    # r"\\": None,
    r"&": None,
}

row_separators = {
    r";": ";",
    r"^\s*$": None,  # Empty line
    r"\\\\": None,
}


fonts = {
    r"\mathrm": {"mathvariant": "normal"},
    r"\mathbf": {"mathvariant": "bold"},
    r"\mathit": {"mathvariant": "italic"},
    r"\mathbit": {"mathvariant": "bold-italic"},

    r"\mathbb": {"mathvariant": "double-struck"},

    r"\mathfrak": {"mathvariant": "fraktur"},
    r"\mathbfrak": {"mathvariant": "bold-fraktur"},

    r"\mathscr": {"mathvariant": "script"},
    r"\mathbscr": {"mathvariant": "bold-script"},

    r"\mathcal": {"mathvariant": "script", "class": "calligraphic"},
    r"\mathbcal": {"mathvariant": "bold-script", "class": "calligraphic"},

    r"\mathsf": {"mathvariant": "sans-serif"},
    r"\mathbsf": {"mathvariant": "bold-sans-serif"},

    r"\mathsfit": {"mathvariant": "sans-serif-italic"},
    r"\mathsfbit": {"mathvariant": "sans-serif-bold-italic"},
    r"\mathbsfit": {"mathvariant": "sans-serif-bold-italic"},  # bold first?

    r"\mathtt": {"mathvariant": "monospace"},

    r"\mathtinit": {"mathvariant": "initial"},
    r"\mathttail": {"mathvariant": "tailed"},
    r"\mathtloop": {"mathvariant": "looped"},
    r"\mathtstre": {"mathvariant": "stretched"},
}


enclosures = {
    r"\cancel": "updiagonalstrike",
    r"\enclose": "roundedbox",
}

spaces = {
    # r"\\ ": &nbsp; &NonBreakingSpace;
    r"\\;": {"width": "3pt"},
	r"\quad": {"width": "1em"},
	r"\thinspace": {"width": "1pt"},
	r"\enspace": {"width": "5pt"},
}


sets = {
	r"\emptyset": "&emptyset;",
	r"\primes": "&Popf;",
	r"\naturals": "&naturals;",
	r"\integers": "&integers;",
	r"\rationals": "&rationals;",
	r"\algebraics": "&Aopf;",
	r"\reals": "&reals;",
	r"\imaginaries": "&Iopf;",
	r"\complexes": "&complexes;",
	r"\quaternions": "&quaternions;",
	r"\octonions": "&Oopf;",
	r"\sedenions": "&Sopf;",

    r"\disc": "𝔻",
    r"\sphere": "𝕊",
}

identifiers.update(sets)

greeks = {
    r"\alpha": "α",  # &alpha;
    r"\beta": "β",  # &beta;
    r"\gamma": "γ",  # &gamma;
    r"\digamma": "ϝ",  # &gammad;
    r"\delta": "δ",  # &delta;
    r"\epsilon": "ϵ",  # &#x3f5;
    r"\varepsilon": "ε",  # &epsilon;
    r"\zeta": "ζ",  # &zeta;
    r"\eta": "η",  # &eta;
    r"\theta": "θ",  # &theta;
    r"\vartheta": "ϑ",  # &#x3D1;
    r"\kappa": "κ",  # &kappa;
    r"\lambda": "λ",  # &lambda;
    r"\mu": "μ",  # &mu;
    r"\nu": "ν",  # &nu;
    r"\xi": "ξ",  # &xi;
    r"\omicron": "ο",  # &omicron;
    r"\pi": "π",  # &pi;
    r"\varpi": "ϖ",  # &#982;
    r"\rho": "ρ",  # &rho;
    r"\varrho": "ϱ",  # &#x3F1;
    r"\sigma": "σ",  # &sigma;
    r"\varsigma": "ς",  # &#x3C2;
    r"\tau": "τ",  # &tau;
    r"\upsilon": "υ",  # &upsilon;
    r"\phi": "ϕ",  # &straightphi;
    r"\varphi": "φ",  # &phi;
    r"\chi": "χ",  # &chi;
    r"\psi": "ψ",  # &psi;
    r"\omega": "ω",  # &omega;
    r"\Alpha": "Α",  # &Alpha;
    r"\Beta": "Β",  # &Beta;
    r"\Gamma": "Γ",  # &Gamma;
    r"\Digamma": "Ϝ",  # &Gammad;
    r"\Delta": "Δ",  # &Delta;
    r"\Zeta": "Ζ",  # &Zeta;
    r"\Eta": "Η",  # &Eta;
    r"\Theta": "Θ",  # &Theta;
    r"\Iota": "Ι",  # &Iota;
    r"\Kappa": "Κ",  # &Kappa;
    r"\Lambda": "Λ",  # &Lambda;
    r"\Mu": "Μ",  # &Mu;
    r"\Nu": "Ν",  # &Nu;
    r"\Xi": "Ξ",  # &Xi;
    r"\Omicron": "Ο",  # &Omicron;
    r"\Pi": "Π",  # &Pi;
    r"\Rho": "Ρ",  # &Rho;
    r"\Sigma": "Σ",  # &Sigma;
    r"\Tau": "Τ",  # &Tau;
    r"\Upsilon": "Υ",  # &Upsilon;
    r"\Phi": "Φ",  # &Phi;
    r"\Chi": "Χ",  # &Chi;
    r"\Psi": "Ψ",  # &Psi;
    r"\Omega": "Ω",  # &Omega;
}

identifiers.update(greeks)



from setuptools import setup, find_packages
from pathlib import Path
from distutils.util import convert_path


# Import __version__ (ugly, but works)
version = Path(__file__).parent / "mumath" / "config.py"
exec(version.read_text(encoding="utf-8"))


long_description = """\
# Mumath
Mumath is a LATEX-like to MathML conversion tool written in Python.

It also includes Markdown extensions for inline `$...$` and blocks `$$...$$`
and a simple server for quick testing.

Install and open in browser:
```sh
$ pip install mumath
$ py -m mumath.server --open
```
"""

packages = find_packages()

options = {
    "name": "mumath",
    "version": __version__,
    "author": "::fourpoints",
    "description": "LATEX to MathML conversion tool",
    "long_description": long_description,
    "long_description_content_type": "text/markdown",
    "url": "https://github.com/fourpoints/mumath",
    "packages": packages,
    # "package_dir": {"": "."},
    "classifiers": [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    "package_data": {
        "mumath": ["data/index.html"],
    },
    "install_requires": [],
    "extras_require": {
        "markdown": ["markdown"],
    },
    "entry_points": {
        "markdown.extensions": [
            "mumath = mumath.markdown:MuMark",
            "mumath-inline = mumath.markdown:InlineMuMark",
        ]
    }
}

setup(**options)



from .context.tokens import *  # builtins.tokenizer does this so why not me
from .node import mml, html, special
from .node.util import clone, traverse
from .util import listify
from itertools import islice, count
from functools import partial, wraps
import xml.etree.ElementTree as ET


# https://github.com/pengbo-learn/python-math-expression-parser

# drawn characters:
# - mi, mn, mo, ms, mspace?, mtext?
# - menclose, mfrac, mroot, msqrt


def diff(tokens, i, attrib=None):
    attrib = attrib if attrib is not None else {}
    # a bit overly optimized, based on the flag value
    # which coincidentally corresponds to the index
    mark = [None, "n", "p", "n p"][tokens[i].flag]

    if mark is not None:
        class_ = attrib.get("class")
        if class_ is None:
            attrib["class"] = mark
        else:
            attrib["class"] = " ".join((class_, mark))
    return attrib


def wrapped(hof):
    # hof = higher order function
    def _hof(func, *args, **kwargs):
        return wraps(func)(hof(func, *args, **kwargs))
    return _hof

partial = wrapped(partial)


def _print(tokens):
    print("".join(tkn.string for tkn in tokens))


def map_get(mapping, key):
    # utility function
    return mapping.property(key, (key, {}))


def nonspace(tokens, i):
    for i, e in enumerate(islice(tokens, i, None), start=i):
        if e.type != SOFT_SPACE: break
    return i


def _string(glyph, tokens, i):
    return tokens[i].string, i+1


def _variant(glyph, tokens, i):
    return tokens[i].variant, i+1


def _isbox(el): return el.tag == "mrow" and len(el) == 1
def _unbox(el): return _unbox(el[0]) if _isbox(el) else el


def number(glyph, tokens, i):
    return mml.mn(tokens[i].string, diff(tokens, i)), i+1

def element(tokens, i, mtype, mapping):
    symbol, attrib = map_get(mapping, tokens[i].variant)
    attrib = diff(tokens, i, attrib)
    return mtype(symbol, attrib), i+1

def identifier(glyph, tokens, i):
    return element(tokens, i, mml.mi, glyph)

def operator(glyph, tokens, i):
    return element(tokens, i, mml.mo, glyph)

def relator(glyph, tokens, i):
    return element(tokens, i, mml.mo, glyph)

def binary_operator(glyph, tokens, i):
    return element(tokens, i, mml.mo, glyph)

def string_literal(glyph, tokens, i):
    return mml.ms(tokens[i].string[1:-1], diff(tokens, i)), i+1

def space(glyph, tokens, i):
    attrib = map_get(glyph, tokens[i].variant)
    return mml.mspace(attrib), i+1

def word(glyph, tokens, i):
    return mml.mi(tokens[i].string), i+1

def comment(glyph, tokens, i):
    el = html.Comment(comment)
    el.text = tokens[i].string[1:].lstrip()
    return el, i+1


def _separator(glyph, tokens, i, ttype, stretch, mapping, attrib):
    i = nonspace(tokens, i)
    if tokens[i].type != ttype:
        raise KeyError
    b, i = _variant(glyph, tokens, i)
    # could _string be used instead?
    if mapping[b].pattern == stretch:
        b, i = _variant(glyph, tokens, i)
        attrib["stretchy"] = "true"
    b = mapping[b].property
    return mml.mo(b, **diff(tokens, i-1, attrib)), i


def open_bracket(glyph, tokens, i, **attrib):
    return _separator(glyph, tokens, i, OPEN, r"\left", glyph, attrib)


def close_bracket(glyph, tokens, i, **attrib):
    return _separator(glyph, tokens, i, CLOSE, r"\right", glyph, attrib)


def middle(glyph, tokens, i, **attrib):
    return _separator(glyph, tokens, i, COL_SEP, r"\middle", glyph, attrib)


def no_number(glyph, tokens, i):
    # Singleton object
    return special.NO_NUMBER, i+1


def text(glyph, tokens, i):
    attrib = diff(tokens, i)
    _b, i = open_bracket(glyph, tokens, i+1)

    def stringed(parts):
        def width(k): return f"{round(k/5, 1)}em"

        n = nonspace(parts, 0)
        m = nonspace(reversed(parts), 0)
        # unfortuntely parts[n:-m] doesn't work for m=0
        text = "".join(e.string for e in parts[n:len(parts)-m])

        strings = []
        if n: strings.append(mml.mspace(width=width(n)))
        strings.append(mml.mtext(text))
        if m: strings.append(mml.mspace(width=width(m)))

        return strings


    mrow = mml.mrow(attrib=attrib)

    parts = []
    while i < len(tokens):
        if tokens[i].type == CLOSE:
            mrow.extend(stringed(parts))
            break
        elif tokens[i].type == TEXT_SEP:
            mrow.extend(stringed(parts))
            parts.clear()
            block, i = blocks(glyph, tokens, i+1)
            mrow.extend(block)

            # block may end before getting to the next $
            while tokens[i].type != TEXT_SEP:
                i += 1
        else:
            parts.append(tokens[i])
        i += 1

    _b, i = close_bracket(glyph, tokens, i)

    return mrow, i


def _subsups(glyph, tokens, i):
    subsup = []
    last = SUP
    while i < len(tokens):
        try:
            ttype = tokens[i].type
            if ttype not in {SUP, SUB}:
                break

            node, i = factor(glyph, tokens, i+1)
            if ttype == last:
                subsup.append(mml.none())
            subsup.append(node)
            last = ttype
        except KeyError:
            break

    if last == SUB:  # Must be even
        subsup.append(mml.none())

    return subsup, i



def _scripts(glyph, tokens, i, ttypes):
    scripts = []

    for ttype in ttypes:
        if i < len(tokens) and tokens[i].type == ttype:
            node, i = factor(glyph, tokens, i+1)
        else:
            node, i = mml.none(), i
        scripts.append(node)

    return scripts, i


_subsup = partial(_scripts, ttypes=[SUB, SUP])
_subbsupp = partial(_scripts, ttypes=[SUBB, SUPP])


def prescripts(glyph, tokens, i):
    mms = mml.mmultiscripts()
    mpre = mml.mprescripts()

    if tokens[i].type == PRESCRIPT:
        i += 1

    prescripts, i = _subsups(glyph, tokens, i)

    try:
        # Add binop or relation here?
        node, i = factor(glyph, tokens, i)
        mms.append(node)
    except KeyError:
        # Technically invalid? Graceful save
        mms.append(mpre)
        mms.extend(prescripts)
        return mms, i

    postscripts, i = _subsups(glyph, tokens, i)

    mms.extend(postscripts)
    mms.append(mpre)
    mms.extend(prescripts)

    return mms, i


def multiscripts(glyph, tokens, i, base):
    postscripts, i = _subsups(glyph, tokens, i)

    if len(postscripts) > 2:
        return mml.mmultiscripts([base, *postscripts]), i
    elif len(postscripts) == 2:
        sub, sup = postscripts
        if base.get("movablelimits"):
            if sub.tag == sup.tag == "none":  # slow
                raise KeyError
            elif sub.tag == "none":
                return mml.mover([base, sup]), i
            elif sup.tag == "none":
                return mml.munder([base, sub]), i
            else:
                return mml.munderover([base, *postscripts]), i
        else:
            if sub.tag == sup.tag == "none":  # slow
                raise KeyError
            elif sub.tag == "none":
                return mml.msup([base, sup]), i
            elif sup.tag == "none":
                return mml.msub([base, sub]), i
            else:
                return mml.msubsup([base, *postscripts]), i
    else:
        raise KeyError


def _binargs(glyph, tokens, i):
    hyper, i = product(glyph, tokens, i)
    base, i = product(glyph, tokens, i)

    return [hyper, base], i


def underset(glyph, tokens, i):
    # generalize overset/underset to hyperscript?
    (under, base), i = _binargs(glyph, tokens, i+1)
    return mml.munder([base, under]), i


def overset(glyph, tokens, i):
    (over, base), i = _binargs(glyph, tokens, i+1)
    return mml.mover([base, over]), i


def underover(glyph, tokens, i, base):
    stack, i = _subbsupp(glyph, tokens, i)
    subb, supp = stack

    if subb.tag == supp.tag == "none":  # slow
        raise KeyError
    elif subb.tag == "none":
        return mml.mover([base, supp]), i
    elif supp.tag == "none":
        return mml.munder([base, subb]), i
    else:
        return mml.munderover([base, *stack]), i


def series(glyph, tokens, i):
    def mutate(node, var, n):
        for el in traverse(node):
            if el.text == var.text:
                el.text = str(n)

    def template(expr, var, n):
        node = clone(expr)
        mutate(node, var, n)
        return node

    def _int(el): return int(el.text)

    (sub, sup), i = _subsup(glyph, tokens, i+1)
    expr, i = product(glyph, tokens, i)
    var, _assignment, start = sub
    end = _unbox(sup)

    start, end = map(_int, (start, end))

    mrow = mml.mrow()

    for n in range(start, end):
        mrow.append(template(expr, var, n))
        mrow.append(mml.mo("+"))
    mrow.append(template(expr, var, end))

    return mrow, i


def group(glyph, tokens, i):
    left, i = open_bracket(glyph, tokens, i, stretchy="false")

    group = []
    while i < len(tokens):
        try:
            block, i = blocks(glyph, tokens, i)
            group.extend(block)
            sep, i = middle(glyph, tokens, i)
            group.append(sep)
        except KeyError:
            break
    right, i = close_bracket(glyph, tokens, i, stretchy="false")

    mrow = mml.mrow()

    # BUGFIX 2023-02-13: we use Text child instead of el.text
    # The fix itself is a bug since it exposes implementation details
    def has_text(xel):
        return len(xel) > 0 and xel[0].tag is str and xel[0].text is not None

    if has_text(left) and has_text(right):
        mrow.extend([left, *group, right])
    elif has_text(left):
        mrow.extend([left, *group])
    elif has_text(right):
        mrow.extend([*group, right])
    else:
        mrow.extend(group)

    return mrow, i


def class_(glyph, tokens, i):
    name, i = factor(glyph, tokens, i+1)
    node, i = factor(glyph, tokens, i)

    name = "".join(name.itertext()).strip()

    node.set("class", name)

    return node, i


def displaystyle(glyph, tokens, i):
    node, i = factor(glyph, tokens, i+1)

    node.set("displaystyle", "true")

    return node, i


def enclose(glyph, tokens, i):
    attrib = diff(tokens, i)
    variant, i = _variant(glyph, tokens, i)
    node, i = factor(glyph, tokens, i)

    notation = glyph[variant].property

    # good?
    if _isbox(node):
        node.tag = "menclose"
        node.set("notation", notation)
    else:
        node = mml.menclose([node], notation=notation)

    return node, i


def pad(glyph, tokens, i):
    node, i = product(glyph, tokens, i+1)
    node = _unbox(node)

    mpadded = mml.mpadded([node], lspace="0.5em", rspace="0.5em")
    return mpadded, i


def function(glyph, tokens, i):
    symbol, attrib = map_get(glyph, tokens[i].variant)
    return mml.mo(symbol, diff(tokens, i, attrib)), i+1


def variant(glyph, tokens, i):
    variant, i = _variant(glyph, tokens, i)
    node, i = factor(glyph, tokens, i)

    style = glyph[variant].property
    fontable = {"mi", "mn", "mo", "ms", "mtext"}

    for el in traverse(node):
        if el.tag in fontable:
            el.attrib.update(style)

    return node, i


def hat(glyph, tokens, i):
    attrib = diff(tokens, i)
    htype, i = _variant(glyph, tokens, i)
    node, i = factor(glyph, tokens, i)
    node = _unbox(node)

    hat = mml.mo(glyph[htype].property, attrib=attrib)

    mover = mml.mover([node, hat], accent="true")
    return mover, i


def shoe(glyph, tokens, i):
    attrib = diff(tokens, i)
    stype, i = _variant(glyph, tokens, i)
    node, i = factor(glyph, tokens, i)
    node = _unbox(node)

    hat = mml.mo(glyph[stype].property, attrib=attrib)

    mover = mml.munder([node, hat], accent="true")
    return mover, i


def norm(glyph, tokens, i):
    attrib = diff(tokens, i)
    ntype, i = _variant(glyph, tokens, i)
    node, i = product(glyph, tokens, i)
    node = _unbox(node)

    def _bracket(b): return mml.mo(b, stretchy="true", attrib=attrib.copy())

    lb, rb = map(_bracket, glyph[ntype].property)

    norm = mml.mrow([lb, node, rb])
    return norm, i


def sqrt(glyph, tokens, i):
    attrib = diff(tokens, i)
    _, i = _variant(glyph, tokens, i)
    node, i = product(glyph, tokens, i)
    node = _unbox(node)

    msqrt = mml.msqrt([node], attrib=attrib)
    return msqrt, i


def root(glyph, tokens, i):
    attrib = diff(tokens, i)
    _, i = _variant(glyph, tokens, i)
    [root, base], i = _binargs(glyph, tokens, i)

    # can be of the form \root[2+2]{3}
    if len(root) > 1 and root[0].text == "[" and root[-1].text == "]":
        # not pretty
        del root[0]
        del root[-1]

    base = _unbox(base)
    root = _unbox(root)

    mroot = mml.mroot([base, root], attrib=attrib)
    return mroot, i


def _sep(glyph, tokens, i, ttype):
    i = nonspace(tokens, i)
    if tokens[i].type == ttype:
        return i + 1
    raise KeyError


_row_sep = partial(_sep, ttype=ROW_SEP)
_col_sep = partial(_sep, ttype=COL_SEP)


def _row(glyph, tokens, i):
    row = []
    while i < len(tokens):
        try:
            cell, i = blocks(glyph, tokens, i)
            row.append(cell)
            i = _col_sep(glyph, tokens, i)
        except KeyError:
            try:
                # in case of empty cell
                i = _col_sep(glyph, tokens, i)
                row.append([])
            except KeyError:
                break

    row = [mml.mtd(cell) for cell in row]
    return row, i


def _table(glyph, tokens, i):
    table = []
    while i < len(tokens):
        try:
            row, i = _row(glyph, tokens, i)
            table.append(row)
            i = _row_sep(glyph, tokens, i)
        except KeyError:
            break

    table = [mml.mtr(row) for row in table]
    return table, i


def matrix(glyph, tokens, i):
    def _brackets(mtype, lbracket, rbracket):
        if mtype == r"\matrix":
            return lbracket, rbracket
        elif mtype == r"\cases":
            return mml.mo("{", stretchy="true"), None
        else:
            return None, None

    mtype, i = _string(glyph, tokens, i)
    lbracket, i = open_bracket(glyph, tokens, i, stretchy="true")
    table, i = _table(glyph, tokens, i)
    rbracket, i = close_bracket(glyph, tokens, i, stretchy="true")

    left, right = _brackets(mtype, lbracket, rbracket)

    bracketed = []
    if left is not None: bracketed.append(left)
    bracketed.append(mml.mtable(table))
    if right is not None: bracketed.append(right)
    matrix = mml.mrow(bracketed)

    return matrix, i

def _environment(glyph, tokens, i):
    lb, i = open_bracket(glyph, tokens, i)
    mvar, i = _variant(glyph, tokens, i)
    rb, i = close_bracket(glyph, tokens, i)
    return mvar, i


def environment(glyph, tokens, i):
    mvar, i = _environment(glyph, tokens, i+1)
    table, i = _table(glyph, tokens, i)
    i = nonspace(tokens, i)
    assert tokens[i].string == r"\end"
    mvar2, i = _environment(glyph, tokens, i+1)
    assert mvar == mvar2

    def _bracket(b): return mml.mo(b, stretchy="true")

    left, right = map(_bracket, glyph[mvar].property)

    bracketed = []
    if left.text is not None: bracketed.append(left)
    bracketed.append(mml.mtable(table))
    if right.text is not None: bracketed.append(right)
    matrix = mml.mrow(bracketed)


    return matrix, i


def factor(glyph, tokens, i):
    type_map = {
        IDENTIFIER: identifier,
        WORD: word,
        # ENVIRONMENT: identifier,
        NUMBER: number,
        OPERATOR: operator,
        SPACE: space,
        SUB: prescripts,
        SUP: prescripts,
        PRESCRIPT: prescripts,
        UNDERSET: underset,
        OVERSET: overset,
        OPEN: group,
        STRING: string_literal,

        FUNCTION: function,
        LARGEOP: operator,
        OPERATOR: operator,
        HAT: hat,
        SHOE: shoe,
        NORM: norm,
        SQRT: sqrt,
        MATRIX: matrix,
        VARIANT: variant,
        ENCLOSE: enclose,
        CLASS_: class_,
        TEXT: text,
        OPEN: group,
        BEGIN: environment,
        SERIES: series,
        NO_NUMBER: no_number,
        COMMENT: comment,

        FRAC: fraction,
        BINOM: binomial,
        ROOT: root,
        DISPLAYSTYLE: displaystyle,
        PAD: pad,

        BINOP: binary_operator,
        RELATION: relator,
    }

    i = nonspace(tokens, i)

    return type_map[tokens[i].type](glyph, tokens, i)


def scripted(glyph, tokens, i, left):
    type_map = {
        SUB: multiscripts,
        SUP: multiscripts,
        SUBB: underover,
        SUPP: underover,
    }

    i = nonspace(tokens, i)

    return type_map[tokens[i].type](glyph, tokens, i, left)


def product(glyph, tokens, i):
    node, i = factor(glyph, tokens, i)

    if i < len(tokens):
        try:
            node, i = scripted(glyph, tokens, i, node)
        except KeyError:
            pass

    return node, i


def is_space(el):
    if el.tag in {"mtext", "mspace", "maligngroup", "malignmark"}:
        return True
    if el.tag in {"mstyle", "mphantom", "mpadded", "mrow"} and all(map(is_space, el)):
        return True
    return False


EMBELLISHABLE_FIRST = {"msub", "msup", "msubsup", "munder", "mover", "munderover", "mmultiscripts", "mfrac"}
EMBELLISHABLE_ANY = {"mstyle", "mphantom", "mpadded", "mrow"}
def is_embellished(el):
    if el.tag == "mo":
        return True
    if el.tag in EMBELLISHABLE_FIRST and is_embellished(el[0]):
        return True
    if el.tag in EMBELLISHABLE_ANY:
        em = sum(map(is_embellished, el))
        sp = sum(map(is_space, el))
        # There can only be one embellished child; the rest must be space-like
        return em == 1 and em + sp == len(el)
    return False


def _embellish(el):
    if el.tag == "mo" and el.get("form"):
        el.set("form", "prefix")
        el.set("rspace", "0")
    elif el.tag in EMBELLISHABLE_FIRST:
        _embellish(el[0])
    elif el.tag in EMBELLISHABLE_ANY:
        em = sum(map(is_embellished, el))
        sp = sum(map(is_space, el))
        # There can only be one embellished child; the rest must be space-like
        if em == 1 and em + sp == len(el):
            _embellish(next(filter(lambda c: c.tag == "mo", el)))
    else:
        # not an operator
        pass


def term(glyph, tokens, i):
    products = []

    prod, i = product(glyph, tokens, i)
    products.append(prod)

    while i < len(tokens):
        try:
            prod, i = product(glyph, tokens, i)

            if prod.tag == "mrow" and prod[0].text == "(":
                _embellish(products[-1])

            # FIXME
            # if products[-1].get("embellished") and prod.get("grouping"):
            #     embellish(products[-1])

            products.append(prod)

        except KeyError:
            break

    return products, i


def binop(glyph, tokens, i):
    i = nonspace(tokens, i)

    if tokens[i].type != BINOP:
        raise KeyError

    binop, i = binary_operator(glyph, tokens, i)

    try:
        binop, i = scripted(glyph, tokens, i, binop)
    except KeyError:
        pass

    return binop, i


def expression(glyph, tokens, i):
    expression = []

    # Check if first element is binary operator
    # If so, make it a unary operator
    try:
        node, i = binop(glyph, tokens, i)
        node.set("prefix", "true")
        expression.append(node)
    except (KeyError, IndexError):
        pass

    while i < len(tokens):
        try:
            terms, i = term(glyph, tokens, i)
            expression.extend(terms)
            try:
                node, i = binop(glyph, tokens, i)
                expression.append(node)
            except (KeyError, IndexError):
                pass
        except KeyError:
            break

    return expression, i


def _collapse(nodes):
    return mml.mrow(nodes) if len(nodes) > 1 else nodes[0]


def _collapse2(left, right):
    if len(left) == len(right) == 1:
        return [left[0], right[0]]
    else:
        return [mml.mrow(left), mml.mrow(right)]


def over(glyph, tokens, i, left):
    # otype, i = _string(glyph, tokens, i)
    attrib = diff(tokens, i)
    right, i = expression(glyph, tokens, i+1)

    frac = mml.mfrac(_collapse2(left, right), attrib=attrib)

    # unsupported
    # if otype == r"\bover":
    #     frac.set("bevelled", "true")

    return [frac], i


def fraction(glyph, tokens, i):
    attrib = diff(tokens, i)
    args, i = _binargs(glyph, tokens, i+1)
    return mml.mfrac(args, attrib=attrib), i


def choose(glyph, tokens, i, left):
    attrib = diff(tokens, i)
    right, i = expression(glyph, tokens, i+1)

    def _bracket(b): return mml.mo(b, attrib=attrib.copy())

    lb, rb = map(_bracket, "()")
    children = _collapse2(left, right)

    mfrac = mml.mfrac(children, linethickness="0", attrib=attrib.copy())

    return [lb, mfrac, rb], i


def binomial(glyph, tokens, i):
    attrib = diff(tokens, i)
    args, i = _binargs(glyph, tokens, i+1)

    def _bracket(b): return mml.mo(b, attrib=attrib.copy())

    lb, rb = map(_bracket, "()")
    children = _collapse2(*args)

    mfrac = mml.mfrac(args, linethickness="0", attrib=attrib.copy())
    mrow = mml.mrow([lb, mfrac, rb])

    return mrow, i


def relation(glyph, tokens, i, left):
    relation, i = relator(glyph, tokens, i)

    try:
        relation, i = scripted(glyph, tokens, i, relation)
    except KeyError:
        pass

    right, i = expression(glyph, tokens, i)

    return [*left, relation, *right], i


def block_operation(glyph, tokens, i, left):
    type_map = {
        OVER: over,
        CHOOSE: choose,
        RELATION: relation,
    }
    i = nonspace(tokens, i)

    return type_map[tokens[i].type](glyph, tokens, i, left)


def blocks(glyph, tokens, i):
    node, i = expression(glyph, tokens, i)

    while i < len(tokens):
        try:
            node, i = block_operation(glyph, tokens, i, node)
        except KeyError:
            break

    return node, i


def _line(glyph, tokens, i):
    row = []
    while i < len(tokens):
        try:
            cell, i = blocks(glyph, tokens, i)
            row.append(cell)
            if i < len(tokens):
                i = _col_sep(glyph, tokens, i)
        except KeyError as e:
            try:
                # in case of empty cell
                i = _col_sep(glyph, tokens, i)
                row.append([])
            except KeyError:
                break

    left = mml.mtd(style="width:50%;padding:0;")
    right = mml.mtd(style="width:50%;padding:0;")
    row = [left] + [mml.mtd(cell) for cell in row] + [right]
    return row, i


def _equation(glyph, tokens, i):
    # almost identical to _table and _row
    table = []
    while i < len(tokens):
        try:
            row, i = _line(glyph, tokens, i)
            table.append(row)
            if i < len(tokens):
                i = _row_sep(glyph, tokens, i)
        except KeyError:
            break

    table = [mml.mtr(row, columnalign="left") for row in table]
    table = mml.mtable(table, displaystyle="true")
    return table, i


def _equation_number(n):
    cell = mml.mtd(style="padding:0;", columnalign="right")
    padding = mml.mpadded(width="2em")
    cell.append(padding)

    if n is not None:
        equation_attrib = dict(id=f"eqn-{n}", href=f"#eqn-{n}")
        padding.attrib.update(equation_attrib)
        number = mml.mtext(f"({n})")
        padding.append(number)

    return cell


def enumeration(table, counter):
    # mutates table
    for row in table:
        count = True
        for cell in row:
            if special.NO_NUMBER in cell:
                # This will only remove one if there are multiple
                # But why would anyone include multiple
                cell.remove(special.NO_NUMBER)
                count = False

        number = next(counter) if count else None

        row.insert(0, _equation_number(number))


def align(glyph, tokens, i):
    table, i = _equation(glyph, tokens, i)
    return [table], i


def inline(glyph, tokens, i):
    node, j = blocks(glyph, tokens, i)
    return node, j


class MathParser:
    _counter =  count(1)

    def __init__(self, glyph, **options):
        self.glyph = glyph
        self.options = options

    def _attrib(self, display):
        # area = self.area
        attrib = {"display": display}

        class_ = f"math math--{display}"
        # if area:
        #     class_ += " ".join(f" marea--{a}" for a in area)
        attrib["class"] = class_

        if display == "block":
            attrib["displaystyle"] = "true"

        return attrib

    def _parse(self, tokens, root, parser):
        nodes, i = parser(self.glyph, tokens, 0)
        root.extend(nodes)

        if 0 == len(nodes):
            nodes.append(mml.merror([mml.mtext("EMPTY EXPRESSION")]))
        elif i < len(tokens):
            nodes.append(mml.merror([mml.mtext("PARSER ERROR")]))

        return root

    def _aligned(self, tokens):
        if self.options.get("infer", False):
            return any(t.string in {r";", r"\\"} for t in tokens)
        else:
            return self.options.get("align", False)

    def parse(self, tokens, root=None):
        aligned = self._aligned(tokens)
        display = self.options.get("display", "block" if aligned else "inline")
        counter = self.options.get("counter")
        parser = align if (display == "block") else inline

        if root is None:
            root = mml.math()

        if isinstance(counter, int) and not isinstance(counter, bool):
            # reset counter
            self._counter = count(int(counter))

        root = self._parse(tokens, root, parser)

        if counter:
            enumeration(root[0], self._counter)

        root.attrib.update(self._attrib(display))

        return ET.ElementTree(root)
